#pragma once

// Defined environment options for KMX processor
#define KM_KBP_KMX_ENV_PLATFORM                       u"platform"
#define KM_KBP_KMX_ENV_BASELAYOUT                     u"baseLayout"
#define KM_KBP_KMX_ENV_BASELAYOUTALT                  u"baseLayoutAlt"
#define KM_KBP_KMX_ENV_SIMULATEALTGR                  u"simulateAltgr"
#define KM_KBP_KMX_ENV_CAPSLOCK                       u"capsLock"
#define KM_KBP_KMX_ENV_BASELAYOUTGIVESCTRLRALTFORRALT u"baseLayoutGivesCtrlRAltForRAlt"
